/***************************************************************************
 *                                                                         *
 *  MAXIMUS-CBCS Source Code, Version 1.02                                 *
 *  Copyright 1989, 1990 by Scott J. Dudley.  All rights reserved.         *
 *                                                                         *
 *  Chat Mode routines (overlayed).  Includes the chat itself, routines    *
 *  for reading IPCxx.BBS, and for processing inbound messages.            *
 *                                                                         *
 *  For complete details of the licensing restrictions, please refer to    *
 *  the licence agreement, which is published in its entirety in MAX.C     *
 *  and LICENCE.MAX.                                                       *
 *                                                                         *
 *  USE OF THIS FILE IS SUBJECT TO THE RESTRICTIONS CONTAINED IN THE       *
 *  MAXIMUS-CBCS LICENSING AGREEMENT.  IF YOU DO NOT FIND THE TEXT OF THIS *
 *  AGREEMENT IN ANY OF THE  AFOREMENTIONED FILES, OR IF YOU DO NOT HAVE   *
 *  THESE FILES, YOU SHOULD IMMEDIATELY CONTACT THE AUTHOR AT ONE OF THE   *
 *  ADDRESSES LISTED BELOW.  IN NO EVENT SHOULD YOU PROCEED TO USE THIS    *
 *  FILE WITHOUT HAVING ACCEPTED THE TERMS OF THE MAXIMUS-CBCS LICENSING   *
 *  AGREEMENT, OR SUCH OTHER AGREEMENT AS YOU ARE ABLE TO REACH WITH THE   *
 *  AUTHOR.                                                                *
 *                                                                         *
 *  You can contact the author at one of the address listed below:         *
 *                                                                         *
 *  Scott Dudley           FidoNet  1:249/106                              *
 *  777 Downing St.        IMEXnet  89:483/202                             *
 *  Kingston, Ont.         Internet f106.n249.z1.fidonet.org               *
 *  Canada - K7M 5N3       BBS      (613) 389-8315 - HST/14.4K, 24hrs      *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <io.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <mem.h>
#include <string.h>
#include "prog.h"
#include "max.h"
#include "alc.h"

extern struct _cstat cstat;

extern int cur_ch,
           num_task,
           *tasks;

extern long scan_time;


static void near pascal ChatNotAvail(int tid);
static void near pascal ChatCantPageSelf(void);

struct _cgs * pascal ChatFindOpen(void)
{
  FFIND *ff;

  static struct _cgs *cgs;

  char temp[PATHLEN];

  int x;


  if (! *PRM(ipc_path))
    return NULL;

  if ((cgs=malloc(sizeof(struct _cgs)))==NULL)
    return NULL;


  /* Scan all of the current IPC files, and read them into an array */

  sprintf(temp,ipc_star,PRM(ipc_path));

  cgs->num_tid=0;

  if ((ff=FindOpen(temp,0))==NULL)
  {
    free(cgs);
    return NULL;
  }

  while (1)
  {
    /* Grab the number out of this one */

    if (sscanf(ff->ff_name,s(ipc_x),&x)==1)
      cgs->tids[cgs->num_tid++]=x;

    if (FindNext(ff) != 0)
      break;
  }

  FindClose(ff);


  /* Now sort them, so they come out in the correct order */

  qksort(cgs->tids,cgs->num_tid);

  /* And 'ptr' tells the FindNext() routine where to start at. */

  cgs->ptr=0;

  return cgs;
}



int pascal ChatFindNext(struct _cgs *cgs,int *tid,char *username,char *status,int *avail)
{
  struct _cstat cstat;
  int fd;

  /* If we're at the end of the line, or if we can't open the IPC, then    *
   * return error.                                                         */

  if (cgs->ptr==cgs->num_tid || (fd=ChatOpenIPC(cgs->tids[cgs->ptr]))==-1)
    return FALSE;

  read(fd,(char *)&cstat,sizeof(struct _cstat));

  ChatCloseIPC(fd);

  /* Now give 'im the info he requested */
  
  if (tid)
    *tid=cgs->tids[cgs->ptr];

  if (username)
    strcpy(username,cstat.username);

  if (status)
    strcpy(status,cstat.status);

  if (avail)
    *avail=cstat.avail;

  /* Inc pointer for next execution */

  cgs->ptr++;

  return TRUE;
}


/* Same as the FindNext() call above, except it checks the stats for a     *
 * particular task number.                                                 */

int pascal ChatFindIndividual(int tid,char *username,char *status,int *avail)
{
  struct _cstat cstat;
  int fd;


  if (! *PRM(ipc_path))
    return FALSE;

  if ((fd=ChatOpenIPC(tid))==-1)
    return FALSE;

  read(fd,(char *)&cstat,sizeof(struct _cstat));

  ChatCloseIPC(fd);

  if (username)
    strcpy(username,cstat.username);

  if (status)
    strcpy(status,cstat.status);

  if (avail)
    *avail=cstat.avail;

  return TRUE;
}


/* Must be called after ChatFindOpen() */

void pascal ChatFindClose(struct _cgs *cgs)
{
  if (cgs)
    free(cgs);
}



/* Cleans up any of the sticky IPC*.BBS files that may be lying around */

void pascal ChatCleanUp(void)
{
  char temp[PATHLEN];

  if (! *PRM(ipc_path))
    return;

  sprintf(temp,ipcxx_bbs,PRM(ipc_path),task_num);
  unlink(temp);
}


void pascal Page(void)
{
  char temp[PATHLEN];

  int avail,
      tid;

  if (! *PRM(ipc_path))
    return;

  for (strcpy(temp,qmark);eqstri(temp,qmark);)
  {
    WhiteN();

    OpusGets(temp,"Please enter the node number to page (?=list): ");

    if (eqstri(temp,qmark))
      Who_Is_On();
  }

  if ((tid=atoi(temp)) != 0)
  {
    if (tid==task_num)
      ChatCantPageSelf();
    else if (! ChatFindIndividual(tid,NULL,NULL,&avail) || !avail)
      ChatNotAvail(tid);
    else
    {
      sprintf(temp,YELLOW "\a\aYou are being paged by " LGREEN "%s" YELLOW " (node %d).",
              usr.name,task_num);

      if (ChatSendMsg(tid,CMSG_page,strlen(temp)+1,temp)==-1)
      #ifdef DEBUG
        Printf("*MsgErr*");
      #else
        ;
      #endif

      Printf(YELLOW "\nWaiting for node %d to respond to page...\n",tid);

      sprintf(linebuf,"%d",tid);
      Chat(FALSE);
    }
  }
}



/* Figure out who is on the other nodes of a multi-line system */

void pascal Who_Is_On(void)
{
  struct _cgs *cgs;

  int tid;

  char nonstop;

  char username[36],
       status[80];


  if (! *PRM(ipc_path))
    return;

  nonstop=FALSE;
  display_line=display_col=1;

  Puts(s(hu_is_on_1));
  Puts(s(hu_is_on_2));

  if ((cgs=ChatFindOpen())==NULL)
    return;

  while (ChatFindNext(cgs,&tid,username,status,NULL))
  {
    Printf(s(hu_is_on_3),username,tid,status,eqstri(username,usr.name) ? " (you)" : blank_str);

    if (MoreYnBreak(&nonstop,CYAN))
      break;

    if (mdm_halt() || brk_trapped)
    {
      brk_trapped=0;
      mdm_dump(DUMP_all);
      break;
    }
  }
  
  ChatFindClose(cgs);
}

void pascal Chat_Toggle(void)
{
  int avail;

  if (! ChatFindIndividual(task_num,NULL,NULL,&avail))
    return;

  if (avail)
    ChatSetStatus(FALSE,s(cs_notavail));
  else ChatSetStatus(TRUE,(s(cs_avail)));
}


void pascal Chat(int use_cb)
{
  struct _css css;
  struct _cgs *cgs;

  char temp[PATHLEN],
       prompt[PATHLEN],
       nextw[21],
       stat[PATHLEN],
       colour;

  int tid,
      avail,
      cc,
      x,y,z;

  *nextw='\0';

  if (! *PRM(ipc_path))
    return;

  if (use_cb)
  {
    WhiteN();

    OpusGets(temp,"Please enter CB channel number (1-255): ");
    cc=atoi(temp);

    if (cc < 1 || cc > 255)
      return;

    sprintf(stat,"Chatting on CB channel %d",cc);
  }
  else
  {
    for (strcpy(temp,qmark);eqstri(temp,qmark);)
    {
      WhiteN();
      
      OpusGets(temp,"Please enter node number to chat with (?=list): ");

      if (eqstri(temp,qmark))
        Who_Is_On();
    }

    x=atoi(temp);

    if (x < 1 || x > 255)
      return;

    /* Make sure that user can only chat with available nodes */
    
    if (x==task_num)
    {
      ChatCantPageSelf();
      return;
    }
    else if (! ChatFindIndividual(x,NULL,NULL,&avail) || !avail)
    {
      ChatNotAvail(x);
      return;
    }

    /* Now calculate a unique channel number.  This number is unique, for  *
     * two nodes talking to each other.                                    */

    y=min(x,task_num);
    z=max(x,task_num);

    cc=(y*256)+z;

    sprintf(stat,"Private chat with node %d",x);
  }

  if (*temp && cc)
  {
    scan_time=0L; /* Force a check to be made */
    Check_For_Message(NULL,NULL);

    cur_ch=cc;

    /* Allocate mem for list of tasks listening to this channel. */

    if ((tasks=malloc(MAX_TASK*sizeof(int)))==NULL)
      return;

    /* Prepare to send a message to all tasks, asking if they're listening *
     * to this channel.                                                    */

    *(int *)temp=cur_ch;
    strcpy(temp+2,usr.name);

    if ((cgs=ChatFindOpen())==NULL)
      return;

    while (ChatFindNext(cgs,&tid,NULL,NULL,&avail))
      if (avail && tid != task_num)
        if (ChatSendMsg(tid,CMSG_enq,strlen(temp+2)+3,temp)==-1)
          #ifdef DEBUG
            Printf("%MsgErr %d%",tid);
          #else
            ;
          #endif

    ChatFindClose(cgs);

    scan_time=CHAT_SCAN_TIME;

    ChatSaveStatus(&css);
    ChatSetStatus(TRUE,stat);

    Puts(WHITE "\nEntering chat mode.  Type `/q' to quit, `/w' to list users on this\n"
                 "channel, and `/?' for help.\n\n");

    while (1)
    {
      colour=(task_num % 6)+9;

      if (colour==10) /* Make sure that colour isn't lightgreen */
        colour=15;

      sprintf(prompt,LGREEN "[%-15.15s]: \x16\x01%c",usr.name,colour);

      if (*nextw)
        strcpy(temp,nextw);
      else *temp='\0';

      if (Input(temp,INPUT_nlb_line | INPUT_nohotflash | INPUT_wordwrap,0,min(usr.width-2,PATHLEN-2)-19,prompt)==1)
        strcpy(nextw,temp+strlen(temp)+1);
      else *nextw='\0';

      if (eqstri(temp,"/w"))
      {
        Puts(YELLOW "\nList of users on this channel:\n\n");

        Puts(WHITE "Name\n");
        Puts(      "\n" YELLOW);

        if (! num_task)
          Printf("You are alone on this channel.\n");

        for (x=0;x < num_task;x++)
          if (ChatFindIndividual(tasks[x],temp,NULL,NULL))
            Printf("%s\n",temp);

        Putc('\n');
      }
      else if (eqstri(temp,"/?"))
        Display_File(0,NULL,ss,PRM(misc_path),"CHATHELP");
      else if (eqstri(temp,"/q"))
        break;
      else    /* Must have been a text string */
      {
        /* Add the text entered to the message */
        strcat(prompt,temp);


        /* Shift the message over one byte, and insert the channel number */
        memmove(prompt+2,prompt,strlen(prompt)+1);
        *(int *)prompt=cur_ch;

        for (x=0;x < num_task;x++)
          if (ChatSendMsg(tasks[x],CMSG_cdata,strlen(prompt+2)+3,prompt)==-1)
          #ifdef DEBUG
            Printf("~MsgErr %d~",tasks[x]);
          #else
            ;
          #endif
      }
    }

    strcpy(temp,usr.name);

    for (x=0;x < num_task;x++)
      if (ChatSendMsg(tasks[x],CMSG_eot,strlen(temp)+1,temp)==-1)
      #ifdef DEBUG
        Printf("@MsgErr %d@",tasks[x]);
      #else
        ;
      #endif

    num_task=0;

    free(tasks);

    ChatRestoreStatus(&css);

    scan_time=0L; /* Force a check to be made */
    Check_For_Message(NULL,NULL);

    scan_time=NORM_SCAN_TIME;
  }

  cur_ch=-1;
}




void pascal Add_To_Chat_List(int tid)
{
  int x;

  #ifdef DEBUG
  Printf("@adding %d to chat list.\n",tid);
  #endif

  if (tasks && num_task < MAX_TASK)
  {
    /* Scan table for task number to add */

    for (x=0;x < num_task;x++)
      if (tasks[x]==tid)
        break;

    /* Don't add task, if it's already there! */

    if (x==num_task)
      tasks[num_task++]=tid;
  }
}




int pascal Remove_From_Chat_List(int tid)
{
  int x,y;

  if (!tasks)
    return FALSE;

  for (x=0;x < num_task;x++)
  {
    if (tasks[x]==tid)
    {
      for (y=x;y < num_task-1;y++)
        tasks[y]=tasks[y+1];

      break;
    }
  }

  if (x==num_task)
    return FALSE;

  num_task--;
  return TRUE;
}

int pascal Header_Chat(int entry)
{
  int x;

  (void)entry;
  
  if (! *linebuf)
  {
    Puts(CLS YELONBLUE CLEOL "The CHAT Section\n");

    Who_Is_On();

    if (ChatFindIndividual(task_num,NULL,NULL,&x))
      Printf(YELLOW "\nYou are currently " LGREEN "%s" YELLOW " for chat.\n\n",
             x ? "AVAILABLE" : "NOT AVAILABLE");
  }

  restart_system=FALSE;
  return TRUE;
}


static void near pascal R_Cleol(void);

static void near pascal R_Cleol(void)
{
  if (graph==GRAPHICS_none)
    Printf("\r\x19 %c\r",current_col);
  else Printf("\r" CLEOL);
}


void pascal ChatHandleMessage(int tid,int type,int len,char *msg,int *tl,int *redo)
{
  char *temp;

  (void)len;

  if (topline && (*tl=topline) != FALSE)
    TopLine(LINE_restore);

  switch (type)
  {
    case CMSG_page:
      /* Space over, instead of CLEOL, for TTY callers */

      R_Cleol();
      Printf(s(ch_page_msg),msg);

      *redo=TRUE;

      if ((temp=malloc(PATHLEN)) != NULL)
      {
        sprintf(temp,"%sCHATPAGE",PRM(misc_path));
        Display_File(0,NULL,temp);
        free(temp);
      }
      break;

    case CMSG_enq:          /* Respond to any incoming ENQs */
      #ifdef DEBUG
      Printf("@enq from tid %d --",tid);
      #endif

      /* If we're hooked up to the requested conference, then ACK it.  *
       * otherwise, ignore.                                            */

      if (*(int *)msg==cur_ch)
      {
        #ifdef DEBUG
        Printf(" sending ack.",tid);
        #endif


        if (ChatSendMsg(tid,CMSG_ack,0,msg)==-1)
        #ifdef DEBUG
          Printf("!MsgErr!");
        #else
          ;
        #endif

        Add_To_Chat_List(tid);

        R_Cleol();

        Printf(s(ch_xx_join),msg+2);

        *redo=TRUE;
      }

      #ifdef DEBUG
      Putc('\n');
      #endif
      break;

    case CMSG_ack:  /* Remote ack'd our ENQ, so add to our chat list. */
      #ifdef DEBUG
      Printf("@ack from tid %d\n",tid);
      #endif

      Add_To_Chat_List(tid);
      break;

    case CMSG_eot:  /* Someone is leaving conversation */
      #ifdef DEBUG
      Printf("@eot from tid %d.\n",tid);
      #endif

      if (Remove_From_Chat_List(tid))
      {
        R_Cleol();
        Printf(s(ch_xx_leave),msg);

        *redo=TRUE;
      }
      break;

    case CMSG_cdata:
      /* If we're not hooked up to this channel, then warn the bimbo   *
       * who sent the message!                                         */

      if (*(int *)msg != cur_ch)
      {
        #ifdef DEBUG
        Printf("#MsgUnSol %d# ",(int)*msg);
        #endif

        ChatSendMsg(tid,CMSG_eot,strlen(usr.name)+1,usr.name);
      }
      else /* just display it! */
      {
        R_Cleol();
        Printf(s(ch_msg),msg+2);
        *redo=TRUE;
      }
      break;
  }
}


    
static void near pascal ChatNotAvail(int tid)
{
  Printf("\nNode %d is not available for chat.\n",tid);
  Press_ENTER();
}

    
static void near pascal ChatCantPageSelf(void)
{
  Puts("\nYou cannot page yourself.\n");
  Press_ENTER();
}
